#include <fastgraf.h>
#include <stdlib.h>
#include <time.h>     // needed for 'randomize'
#include "stats.h"
#include "pit.h"
#include "game.h"


pit_grid grid;

CUELIST(Pit)
  KEYSTROKE(ESC,on_escape)
  KEYSTROKE('5',on_rotate)
  KEYSTROKE('4',on_left)
  KEYSTROKE('6',on_right)
  KEYSTROKE('2',on_down)
  KEYSTROKE('0',on_drop)
  MESSAGE(NEXT_LEVEL,on_next_level)
  MESSAGE(GAME_OVER,on_game_over)
ENDCUELIST

Pit::Pit()
  {
  jadeblock.set_director(this);
  blueblock.set_director(this);
  grayblock.set_director(this);
  pinkblock.set_director(this);
  yellblock.set_director(this);
  smokblock.set_director(this);
  turqblock.set_director(this);
  }

void Pit::initialize()
  {
  statsptr=new Statistics(this,"font.gfx");
  randomize();
  }

void Pit::display()
  {
  gravity_flag=TRUE;
  for (int row=0;row<PIT_DEPTH;row++)
    {
    for (int col=0;col<PIT_WIDTH;col++)
      grid.row[row].col[col]=NULL;
    grid.row[row].is_full=FALSE;
    }

  init_video();
  show_pcx("bg.pcx");
  swap_video_pages();
  synch_video_pages();
  fill_background_buffer(0);
  statsptr->show();

  blockcount=0;
  level=1;

  reset_current_block();
  update_screen();
  request_timer_cue(level,(callback)&Pit::on_timer);
  }

void Pit::hide()
  {
  stop_timer_cue(level,(callback)&Pit::on_timer);
  }

void Pit::on_escape(int)
  {
  stop_director();
  }

void Pit::on_rotate(int)
  {
  if (current_block->can_rotate())
    {
    current_block->rotate();
    update_screen();
    }
  }

void Pit::on_left(int)
  {
  if (current_block->can_move(LEFT))
    {
    current_block->move(LEFT);
    update_screen();
    }
  }

void Pit::on_right(int)
  {
  if (current_block->can_move(RIGHT))
    {
    current_block->move(RIGHT);
    update_screen();
    }
  }

void Pit::on_down(int)
  {
  if (current_block->can_move(DOWN))
    {
    current_block->move(DOWN);
    update_screen();
    gravity_flag=FALSE;
    }
  else
    {
    if (some_rows_are_complete())
      {
      current_block->play_clip(ROWREMOVAL);
      delete_complete_rows();
      update_screen_pit();
      }
    else
      current_block->play_clip(BLOCKATTACH);
    reset_current_block();
    update_screen();
    }
  }

void Pit::on_drop(int)
  {
  current_block->drop();
  update_screen();
  if (some_rows_are_complete())
    {
    current_block->play_clip(ROWREMOVAL);
    fg_waitfor(1);
    delete_complete_rows();
    }
  else
    current_block->play_clip(BLOCKATTACH);
  update_screen_pit();
  reset_current_block();
  update_screen();
  }

void Pit::on_timer()
  {
  if (!gravity_flag)
    {
    gravity_flag=TRUE;
    return;
    }
  if (current_block->can_move(DOWN))
    {
    current_block->move(DOWN);
    update_screen();
    }
  else
    {
    if (some_rows_are_complete())
      {
      current_block->play_clip(ROWREMOVAL);
      delete_complete_rows();
      update_screen_pit();
      }
    else
      current_block->play_clip(BLOCKATTACH);
    reset_current_block();
    update_screen();
    }
  }

void Pit::on_next_level()
  {
  stop_timer_cue(level-1,(callback)&Pit::on_timer);
  request_timer_cue(level,(callback)&Pit::on_timer);
  }

void Pit::on_game_over()
  {
  statsptr->show_game_over();
  swap_video_pages();
  fg_waitkey();
  stop_director();
  }

void Pit::reset_current_block()
  {
  if (++blockcount >= BLOCK_COUNT)
    {
    post_message(NEXT_LEVEL,++level);
    blockcount=0;
    }
  static int oldr;
  int r=(rand()>>4)%7;
  if (r==oldr)  r=(rand()>>4)%7;  // reduce repeats
  oldr=r;

  if (r==0)
    current_block=&jadeblock;
  else if (r==1)
    current_block=&blueblock;
  else if (r==2)
    current_block=&grayblock;
  else if (r==3)
    current_block=&pinkblock;
  else if (r==4)
    current_block=&yellblock;
  else if (r==5)
    current_block=&smokblock;
  else if (r==6)
    current_block=&turqblock;

  current_block->reset();
  }

void Pit::update_screen()
  {
  current_block->erase();
  current_block->draw();
  swap_video_pages();
  synch_patches();
  }

void Pit::update_screen_pit()
  {
  int x,y;
  restore_patch(PIT_EDGE,0,PIT_EDGE+159,239);
  for (int row=0;row<PIT_DEPTH;row++)
    {
    for (int col=0;col<PIT_WIDTH;col++)
      if (grid.row[row].col[col]!=NULL)
        {
        x=PIT_EDGE+col*16;
        y=row*16;
        grid.row[row].col[col]->draw(x,y);
        }
    }
  swap_video_pages();
  synch_patch(PIT_EDGE,0,PIT_EDGE+159,239);
  }

int Pit::some_rows_are_complete()
  {
  int ret=FALSE;
  for (int row=0;row<PIT_DEPTH;row++)
    {
    int found_blank=FALSE;
    for (int col=0;col<PIT_WIDTH && !found_blank;col++)
      {
      if (grid.row[row].col[col]==NULL)  found_blank=TRUE;
      }
    if (!found_blank)
      {
      grid.row[row].is_full=TRUE;
      ret=TRUE;
      }
    }
  return ret;
  }

void Pit::delete_complete_rows()
  {
  int nrows=0;
  for (int row=PIT_DEPTH-1;row>0;row--)
    {
    if (grid.row[row].is_full)
      {
      shift_down_at_row(row);
      row++;
      nrows++;
      }
    }
  post_message(ROWS_REMOVED,nrows);
  }

void Pit::shift_down_at_row(int row)
  {
  int r,c;
  for (r=row;r>0;r--)
    {
    for (c=0;c<PIT_WIDTH;c++)
      grid.row[r].col[c]=grid.row[r-1].col[c];
    grid.row[r].is_full=grid.row[r-1].is_full;
    }
  for (c=0;c<PIT_WIDTH;c++)
    grid.row[0].col[c]=NULL;
  grid.row[0].is_full=FALSE;
  }

void Pit::take_over()
  {
  fg_setnum(ON);
  Director::take_over();
  fg_setnum(OFF);
  }

//---------------------------------------- blocks ------------

void Block::set()
  {
  int i=0;
  int ox,oy;
  pointlist& list=listptr[rot];

  ox=list.pair[i].x;
  oy=list.pair[i++].y;
  if (y+oy >= 0)
    grid.row[y+oy].col[x+ox]=this;

  ox=list.pair[i].x;
  oy=list.pair[i++].y;
  if (y+oy >= 0)
    grid.row[y+oy].col[x+ox]=this;

  ox=list.pair[i].x;
  oy=list.pair[i++].y;
  if (y+oy >= 0)
    grid.row[y+oy].col[x+ox]=this;

  ox=list.pair[i].x;
  oy=list.pair[i++].y;
  if (y+oy >= 0)
    grid.row[y+oy].col[x+ox]=this;
  }

void Block::unset()
  {
  int i=0;
  int ox,oy;
  pointlist& list=listptr[rot];

  ox=list.pair[i].x;
  oy=list.pair[i++].y;
  if (y+oy >= 0)
    grid.row[y+oy].col[x+ox]=NULL;

  ox=list.pair[i].x;
  oy=list.pair[i++].y;
  if (y+oy >= 0)
    grid.row[y+oy].col[x+ox]=NULL;

  ox=list.pair[i].x;
  oy=list.pair[i++].y;
  if (y+oy >= 0)
    grid.row[y+oy].col[x+ox]=NULL;

  ox=list.pair[i].x;
  oy=list.pair[i++].y;
  if (y+oy >= 0)
    grid.row[y+oy].col[x+ox]=NULL;
  }

int Block::can_set(int x,int y)
  {
  int i=0;
  int ox,oy;
  pointlist& list=listptr[rot];

  ox=list.pair[i].x;
  oy=list.pair[i++].y;
  if (y+oy >= PIT_DEPTH) return FALSE;
  if (x+ox < 0 || x+ox >= PIT_WIDTH)  return FALSE;
  if (y+oy>=0 && grid.row[y+oy].col[x+ox])  return FALSE;

  ox=list.pair[i].x;
  oy=list.pair[i++].y;
  if (y+oy >= PIT_DEPTH) return FALSE;
  if (x+ox < 0 || x+ox >= PIT_WIDTH)  return FALSE;
  if (y+oy>=0 && grid.row[y+oy].col[x+ox])  return FALSE;

  ox=list.pair[i].x;
  oy=list.pair[i++].y;
  if (y+oy >= PIT_DEPTH) return FALSE;
  if (x+ox < 0 || x+ox >= PIT_WIDTH)  return FALSE;
  if (y+oy>=0 && grid.row[y+oy].col[x+ox])  return FALSE;

  ox=list.pair[i].x;
  oy=list.pair[i++].y;
  if (y+oy >= PIT_DEPTH) return FALSE;
  if (x+ox < 0 || x+ox >= PIT_WIDTH)  return FALSE;
  if (y+oy>=0 && grid.row[y+oy].col[x+ox])  return FALSE;

  return TRUE;
  }

int Block::can_set(int r)
  {
  int i=0;
  int ox,oy;
  pointlist& list=listptr[r];

  ox=list.pair[i].x;
  oy=list.pair[i++].y;
  if (y+oy>=0)
    {
    if (x+ox < 0 || x+ox >= PIT_WIDTH)  return FALSE;
    if (grid.row[y+oy].col[x+ox])  return FALSE;
    }
  ox=list.pair[i].x;
  oy=list.pair[i++].y;
  if (y+oy>=0)
    {
    if (x+ox < 0 || x+ox >= PIT_WIDTH)  return FALSE;
    if (grid.row[y+oy].col[x+ox])  return FALSE;
    }
  ox=list.pair[i].x;
  oy=list.pair[i++].y;
  if (y+oy>=0)
    {
    if (x+ox < 0 || x+ox >= PIT_WIDTH)  return FALSE;
    if (grid.row[y+oy].col[x+ox])  return FALSE;
    }
  ox=list.pair[i].x;
  oy=list.pair[i++].y;
  if (y+oy>=0)
    {
    if (x+ox < 0 || x+ox >= PIT_WIDTH)  return FALSE;
    if (grid.row[y+oy].col[x+ox])  return FALSE;
    }
  return TRUE;
  }

void Block::draw()
  {
  int i=0;
  int col,row;
  int sx,sy;
  pointlist& list=listptr[rot];

  col=x+list.pair[i].x;
  row=y+list.pair[i++].y;
  sx=col*16+PIT_EDGE;
  sy=row*16;

  if (row>=0)
    {
    show_image(sx,sy,image);
    VideoDirector::set_synch_patch(sx,sy,sx+15,sy+15);
    }

  col=x+list.pair[i].x;
  row=y+list.pair[i++].y;
  sx=col*16+PIT_EDGE;
  sy=row*16;
  if (row>=0)
    {
    show_image(sx,sy,image);
    VideoDirector::set_synch_patch(sx,sy,sx+15,sy+15);
    }

  col=x+list.pair[i].x;
  row=y+list.pair[i++].y;
  sx=col*16+PIT_EDGE;
  sy=row*16;
  if (row>=0)
    {
    show_image(sx,sy,image);
    VideoDirector::set_synch_patch(sx,sy,sx+15,sy+15);
    }

  col=x+list.pair[i].x;
  row=y+list.pair[i++].y;
  sx=col*16+PIT_EDGE;
  sy=row*16;
  if (row>=0)
    {
    show_image(sx,sy,image);
    VideoDirector::set_synch_patch(sx,sy,sx+15,sy+15);
    }
  }

void Block::erase()
  {
  int i=0;
  int col,row;
  int sx,sy;
  pointlist& list=listptr[oldrot];

  col=ox+list.pair[i].x;
  row=oy+list.pair[i++].y;
  sx=col*16+PIT_EDGE;
  sy=row*16;
  if (row>=0)
    {
    VideoDirector::restore_patch(sx,sy,sx+15,sy+15);
    VideoDirector::set_synch_patch(sx,sy,sx+15,sy+15);
    }

  col=ox+list.pair[i].x;
  row=oy+list.pair[i++].y;
  sx=col*16+PIT_EDGE;
  sy=row*16;
  if (row>=0)
    {
    VideoDirector::restore_patch(sx,sy,sx+15,sy+15);
    VideoDirector::set_synch_patch(sx,sy,sx+15,sy+15);
    }

  col=ox+list.pair[i].x;
  row=oy+list.pair[i++].y;
  sx=col*16+PIT_EDGE;
  sy=row*16;
  if (row>=0)
    {
    VideoDirector::restore_patch(sx,sy,sx+15,sy+15);
    VideoDirector::set_synch_patch(sx,sy,sx+15,sy+15);
    }

  col=ox+list.pair[i].x;
  row=oy+list.pair[i++].y;
  sx=col*16+PIT_EDGE;
  sy=row*16;
  if (row>=0)
    {
    VideoDirector::restore_patch(sx,sy,sx+15,sy+15);
    VideoDirector::set_synch_patch(sx,sy,sx+15,sy+15);
    }
  }
